#define _WIN32_WINNT 0x0501
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shlwapi.h>

#ifndef ABM_SETSTATE
#define ABM_SETSTATE 0x0000000a
#endif // #ifndef ABM_SETSTATE

typedef EXECUTION_STATE (WINAPI *SETTHREADEXECUTIONSTATEPROC)(EXECUTION_STATE);

static const int HIDE_TASKBAR       = 0x01;
static const int RESIZE_SCREEN      = 0x02;
static const int DISABLE_SONAR      = 0x04;
static const int KEEP_DISPLAY_ON    = 0x08;
static bool                         _bScreenResolutionChanged;
static BOOL                         _bSonar;
static bool                         _bAppBarStatusChanged;
static UINT_PTR                     _uState;
static SETTHREADEXECUTIONSTATEPROC  _fnSetThreadExecutionState;
static BOOL                         _bScreenSaverActive;

static DECLSPEC_NORETURN void
ShowMessageAndExit(LPCSTR lpText, UINT uType, UINT uExitCode)
{
    MessageBoxA(NULL, lpText, "Foscre", uType);
    ExitProcess(uExitCode);
}

static DECLSPEC_NORETURN void ShowErrorAndExit(LPCSTR lpText)
{ ShowMessageAndExit(lpText, MB_ICONERROR, 1); }

static DECLSPEC_NORETURN void SyntaxError()
{
    ShowMessageAndExit(
        "Syntax:\n"
        "FOSCRE options file [parametes]\n\n"
        "Available options:\n"
        "  /R:width\x0d7height\tSwitches to a specified screen resolution.\n"
        "  /S\t\tDisables Mouse Sonar (requires Windows XP or later).\n"
        "  /T\t\tAutomatically hides the taskbar (requires Windows XP or "
        "later).\n"
        "  /U\t\tKeeps the display(s) on (requires Windows 2000 or Windows XP "
        "or later).\n"
        "\n"
        "Example:\n"
        "foscre /r:640x480 charmap\n"
        "This will run the character map in 640\x0d7""480 screen resolution.\n",
        MB_ICONINFORMATION,
        0
        );
}

static DWORD NewPels(WCHAR ch, DWORD dwPels)
{
    if (ch < '0' || ch > '9' || dwPels >= 10000) SyntaxError();
    return 10 * dwPels + ch - '0';
}

static int AddSimpleOption(int options, int newOption, const WCHAR *pch)
{
    if (options & newOption || *pch != ' ') SyntaxError();
    return options | newOption;
}

static BOOL SetScreenSaverActive(BOOL bActive)
{ return SystemParametersInfoA(SPI_SETSCREENSAVEACTIVE, bActive, NULL, 0); }

static BOOL SystemParametersInfo(UINT uiAction, PVOID pvParam)
{ return SystemParametersInfoA(uiAction, 0, pvParam, 0); }

static LONG ChangeScreenResolution(LPDEVMODEA lpDevMode)
{ return ChangeDisplaySettingsA(lpDevMode, CDS_FULLSCREEN); }

static void ResetScreenResolution()
{ ChangeScreenResolution(NULL); }

static void Undo()
{
    {
        SETTHREADEXECUTIONSTATEPROC
            fnSetThreadExecutionState = _fnSetThreadExecutionState;
        if (fnSetThreadExecutionState != NULL)
        {
            if (_bScreenSaverActive) SetScreenSaverActive(TRUE);
            fnSetThreadExecutionState(ES_CONTINUOUS);
        }
    }
    if (_bAppBarStatusChanged)
    {
        APPBARDATA abd;
        abd.cbSize = sizeof abd;
        abd.lParam = _uState;
        SHAppBarMessage(ABM_SETSTATE, &abd);
    }
    if (_bSonar) SystemParametersInfo(SPI_SETMOUSESONAR, (PVOID)TRUE);
    if (_bScreenResolutionChanged) ResetScreenResolution();
}

static DECLSPEC_NORETURN void ShowErrorAndExit(HANDLE hMutex, LPCSTR lpText)
{
    Undo();
    CloseHandle(hMutex);
    ShowErrorAndExit(lpText);
}

static FARPROC GetProcAddress(LPCSTR lpModuleName, LPCSTR lpProcName)
{ return GetProcAddress(GetModuleHandleA(lpModuleName), lpProcName); }

static FARPROC GetShell32ProcAddress(LPCSTR lpProcName)
{ return GetProcAddress("SHELL32.DLL", lpProcName); }

static void HideTaskbar(HANDLE hMutex)
{
    FARPROC fnDllGetVersion = GetShell32ProcAddress("DllGetVersion");
    if (fnDllGetVersion != NULL)
    {
        DLLVERSIONINFO dvi;
        dvi.cbSize = sizeof dvi;
        if (
            ((HRESULT (CALLBACK *)(DLLVERSIONINFO *))fnDllGetVersion)(&dvi) ==
            NOERROR &&
            dvi.dwMajorVersion >= 6)
        {
            APPBARDATA abd;
            abd.cbSize = sizeof abd;
            abd.lParam =
                (_uState = SHAppBarMessage(ABM_GETSTATE, &abd)) | ABS_AUTOHIDE;
            SHAppBarMessage(ABM_SETSTATE, &abd);
            if (SHAppBarMessage(ABM_GETSTATE, &abd) & ABS_AUTOHIDE)
            {
                _bAppBarStatusChanged = true;
                return;
            }
        }
    }
    ShowErrorAndExit(hMutex, "Cannot hide the taskbar.");
}

static DECLSPEC_NORETURN void ShowOpenFailedErrorAndExit(HANDLE hMutex)
{ ShowErrorAndExit(hMutex, "Failed to open the specified item."); }

static DECLSPEC_NORETURN void UndoAndExit()
{
    Undo();
    ExitProcess(0);
}

static LRESULT
CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    if (uMsg == WM_ENDSESSION) UndoAndExit();
    return DefWindowProcA(hwnd, uMsg, wParam, lParam);
}

extern "C" DECLSPEC_NORETURN void WINAPI EntryPoint()
{
    if ((GetVersion() & 0xff) < 4)
        ShowErrorAndExit("Operating system not supported.");
    int options = 0;
    DWORD dwPelsWidth = 0, dwPelsHeight = 0;
    LPCWSTR lpFile;
    LPWSTR pch = GetCommandLineW();
    {
        bool bQuotOpen = false;
        bool bWordFound = false;
        for (;;)
        {
            WCHAR ch = *pch++;
            if (ch == '\0') SyntaxError();
            if (ch == ' ')
            {
                if (bWordFound && !bQuotOpen) break;
            }
            else
            {
                bWordFound = true;
                if (ch == '\"') bQuotOpen = !bQuotOpen;
            }
        }
    }
    {
        bool bQuotOpen;
        for (;; ++pch)
        {
            WCHAR ch = *pch;
            if (ch == '-' || ch == '/')
            {
                WCHAR ch = *++pch;
                if (ch == 'r' || ch == 'R')
                {
                    if (options & RESIZE_SCREEN) SyntaxError();
                    WCHAR ch = *++pch;
                    if (ch != ':') SyntaxError();
                    for (bool bDigitFound = false;; bDigitFound = true)
                    {
                        WCHAR ch = *++pch;
                        if (ch == 'x' || ch == 'X' || ch == L'\x0d7')
                        {
                            if (!bDigitFound) SyntaxError();
                            break;
                        }
                        dwPelsWidth = NewPels(ch, dwPelsWidth);
                    }
                    for (bool bDigitFound = false;; bDigitFound = true)
                    {
                        WCHAR ch = *++pch;
                        if (ch == ' ')
                        {
                            if (!bDigitFound) SyntaxError();
                            break;
                        }
                        dwPelsHeight = NewPels(ch, dwPelsHeight);
                    }
                    options |= RESIZE_SCREEN;
                }
                else if (ch == 's' || ch == 'S')
                    options = AddSimpleOption(options, DISABLE_SONAR, ++pch);
                else if (ch == 't' || ch == 'T')
                    options = AddSimpleOption(options, HIDE_TASKBAR, ++pch);
                else if (ch == 'u' || ch == 'U')
                    options = AddSimpleOption(options, KEEP_DISPLAY_ON, ++pch);
                else SyntaxError();
            }
            else if (ch != ' ')
            {
                if (options == 0) SyntaxError();
                bQuotOpen = ch == '\"';
                break;
            }
        }
        lpFile = pch;
        for (;;)
        {
            WCHAR ch = *++pch;
            if (ch == '\"') bQuotOpen = !bQuotOpen;
            else if (ch == '\0') break;
            else if (ch == ' ')
            {
                if (!bQuotOpen)
                {
                    *pch = '\0';
                    while (*++pch == ' ');
                    break;
                }
            }
        }
    }
    HANDLE hMutex = CreateMutexA(NULL, TRUE, "FOSCRE");
    if (GetLastError() == ERROR_SUCCESS)
    {
        if (options & RESIZE_SCREEN)
        {
            {
                // The registry value contained in
                // HKEY_LOCAL_MACHINE\Config\0001\Display\Settings\BitsPerPixel
                // could be invalid, which would prevent ChangeDisplaySettings
                // from succeeding in both changing and restoring the screen
                // resolution as the color depth is not explicitly specified.
                // The first call to ChangeDisplaySettingsA will work this
                // around on Windows 98 and Me (Windows 95 does not support the
                // flag combination of CDS_NORESET and CDS_UPDATEREGISTRY which
                // is required not to mess up with the rendering).
                DEVMODEA devMode;
                devMode.dmSize          = sizeof devMode;
                devMode.dmDriverExtra   = 0;
                devMode.dmFields        = DM_BITSPERPEL;
                devMode.dmBitsPerPel    = GetDeviceCaps(GetDC(NULL), BITSPIXEL);
                ChangeDisplaySettingsA(
                    &devMode,
                    CDS_NORESET | CDS_UPDATEREGISTRY
                    );
                devMode.dmFields        = DM_PELSWIDTH | DM_PELSHEIGHT;
                devMode.dmPelsWidth     = dwPelsWidth;
                devMode.dmPelsHeight    = dwPelsHeight;
                if (ChangeScreenResolution(&devMode) != DISP_CHANGE_SUCCESSFUL)
                    ShowErrorAndExit(
                        hMutex,
                        "Cannot switch to the requested screen resolution."
                        );
            }
            {
                HWND hwnd = FindWindowA("Shell_TrayWnd", "");
                // Under 2000 and Me the tray bar is not repositioned
                // automatically before processing this message.
                SendMessageTimeoutA(
                    hwnd,
                    WM_SETTINGCHANGE,
                    0,
                    0,
                    SMTO_ABORTIFHUNG,
                    2000,
                    NULL
                    );
                // Windows 98 additionally requires this call to get the tray
                // bar correctly repositioned.
                SetWindowPos(
                    hwnd,
                    NULL,
                    0,
                    0,
                    0,
                    0,
                    SWP_ASYNCWINDOWPOS |
                    SWP_DEFERERASE |
                    SWP_NOACTIVATE |
                    SWP_NOMOVE |
                    SWP_NOSIZE |
                    SWP_NOZORDER
                    );
            }
            _bScreenResolutionChanged = true;
        }
        if (options & DISABLE_SONAR)
        {
            if (
                !SystemParametersInfo(SPI_GETMOUSESONAR, &_bSonar) ||
                _bSonar &&
                !SystemParametersInfo(SPI_SETMOUSESONAR, (PVOID)FALSE))
                ShowErrorAndExit(hMutex, "Cannot disable Mouse Sonar.");
        }
        if (options & HIDE_TASKBAR) HideTaskbar(hMutex);
        if (options & KEEP_DISPLAY_ON)
        {
            SETTHREADEXECUTIONSTATEPROC fnSetThreadExecutionState =
                (SETTHREADEXECUTIONSTATEPROC)
                GetProcAddress("KERNEL32.DLL", "SetThreadExecutionState");
            if (
                fnSetThreadExecutionState == NULL ||
                fnSetThreadExecutionState(ES_CONTINUOUS | ES_DISPLAY_REQUIRED)
                ==
                0)
                ShowErrorAndExit(
                hMutex,
                "Cannot change monitor power settings."
                );
            _fnSetThreadExecutionState = fnSetThreadExecutionState;
            if (
                !SystemParametersInfo(
                SPI_GETSCREENSAVEACTIVE,
                &_bScreenSaverActive)
                ||
                _bScreenSaverActive && !SetScreenSaverActive(FALSE))
                ShowErrorAndExit(hMutex, "Cannot disable the screen saver.");
        }
    }
    else // Another instance of Foscre is already running.
        hMutex = NULL;
    {
        SHELLEXECUTEINFOW sei;
        sei.cbSize          = sizeof sei;
        sei.fMask           = SEE_MASK_DOENVSUBST | SEE_MASK_NOCLOSEPROCESS;
        sei.hwnd            = NULL;
        sei.lpVerb          = NULL;
        sei.lpFile          = lpFile;
        sei.lpParameters    = pch;
        sei.lpDirectory     = NULL;
        sei.nShow           = SW_SHOWNORMAL;
        if (GetVersion() & 0x80000000)
        {
            // Convert lpFile to ANSI
            for (size_t index = 0;; ++index)
            {
                WCHAR ch = lpFile[index];
                if (ch & 0xff00) ShowOpenFailedErrorAndExit(hMutex);
                ((CHAR *)lpFile)[index] = (CHAR)ch;
                if (ch == '\0')
                    break;
            }
            // Convert lpParameters to ANSI
            for (size_t index = 0;; ++index)
            {
                WCHAR ch = pch[index];
                if (ch & 0xff00) ShowOpenFailedErrorAndExit(hMutex);
                ((CHAR *)pch)[index] = (CHAR)ch;
                if (ch == '\0')
                {
                    ShellExecuteExA((LPSHELLEXECUTEINFOA)&sei);
                    break;
                }
            }
        }
        else
        {
            FARPROC fn = GetShell32ProcAddress("ShellExecuteExW");
            if (fn == NULL) ShowOpenFailedErrorAndExit(hMutex);
            ((BOOL (STDAPICALLTYPE *)(LPSHELLEXECUTEINFOW))fn)(&sei);
        }
        if (hMutex != NULL && sei.hProcess != NULL)
        {
            LPCSTR szClassName = "Foscre";
            {
                WNDCLASSA wndClass;
                wndClass.style          = 0;
                wndClass.lpfnWndProc    = WndProc;
                wndClass.cbClsExtra     = 0;
                wndClass.cbWndExtra     = 0;
                wndClass.hInstance      = NULL;
                wndClass.hIcon          = NULL;
                wndClass.hCursor        = NULL;
                wndClass.hbrBackground  = NULL;
                wndClass.lpszMenuName   = NULL;
                wndClass.lpszClassName  = szClassName;
                RegisterClassA(&wndClass);
            }
            if (
                CreateWindowExA(
                0,
                szClassName,
                NULL,
                0,
                0,
                0,
                0,
                0,
                NULL,
                NULL,
                NULL,
                NULL)
                != NULL)
                while (
                    MsgWaitForMultipleObjects(
                    1,
                    &sei.hProcess,
                    FALSE,
                    INFINITE,
                    QS_ALLINPUT)
                    ==
                    WAIT_OBJECT_0 + 1)
                {
                    MSG msg;
                    PeekMessageA(&msg, NULL, 0, 0, PM_REMOVE);
                }
        }
    }
    UndoAndExit();
}
